local base = _G

module('me_options_assignment')

local i18n = base.require('i18n')
local form = base.require('me_options_assignment_form')
local Options = base.require('me_options')
local NewInput = base.require('NewInput')
local U = base.require('me_utilities')

i18n.setup(_M)

cdata = {
    add_assignment_panel = _('ADD ASSIGNMENT PANEL'),
    action = _('Action:'),
    key_button = _('Key, Button'),
    add_modifier = _('Add Modifier'),
    added_modifiers = _('Added Modifiers'),
    currently_in_use = _('Currently in Use'),
    ok = _('OK'),
    cancel = _('CANCEL'),
    reset = _('RESET'),
    axisCommands = _("Axis Commands"),
};

function create(x, y, w, h)
    window = form.create(x, y, w, h, cdata)
    setupCallbacks();
    return window;
end

function show(b)
    local activeCell = Options.activeCell;
    local category = Options.form.containerControls.comboboxCategory:getText();
    local keyLayerName;
    local axesLayerName;
    DEVICE = nil;
    if b then 
        Options.setWindowGray(true);
        if activeCell then
            local action = activeCell.data.action;
            local deviceDescription = action.device;
            DEVICE = deviceDescription;
            window.controls.editboxAction:setText(action.localizedName);
            if cdata.axisCommands == category then
                U.fill_combo(window.controls.comboboxKeyButton, 
                    U.getSortedKeysTable(deviceDescription.supportedAxes));
                axesLayerName = createAxesLayer();
            else
                U.fill_combo(window.controls.comboboxKeyButton, 
                    U.getSortedKeysTable(deviceDescription.supportedKeys));
                keyLayerName = createKeyLayer();
            end;
            window.controls.comboboxKeyButton:setText('');
            local t = {};
            -- for k,v in base.pairs(Options.ALL_REFORMERS_CODES) do
                -- base.table.insert(t,v);
            for k,v in base.pairs(Options.ALL_REFORMERS) do
                base.table.insert(t,k);
            end;
            base.table.sort(t);
            U.fill_combo(window.controls.comboboxAddModifier, t);
            window.controls.comboboxAddModifier:setText('');
            window.controls.editboxAddedModifiers:setText('');
            window.controls.editboxCurrentSetting:setText('');
        end;
        
        window:setTop(true);
        window:setModal(true);
        window:setVisible(true);
    else 
        base.setInputProcessor(nil);
        hideWindow();
    end;
    --base.print('show',b);
end; 




function setupCallbacks()
     window.controls.buttonOk.onChange = onButtonOk;
     window.controls.buttonReset.onChange = onButtonReset;
     window.controls.buttonCancel.onChange = onButtonCancel;
     window.controls.comboboxAddModifier.onChange = onComboAddModifier;
     window.controls.comboboxKeyButton.onChange = onComboKeyButton;
     window.onClose = function(self) hideWindow(); end;
end; 

function hideWindow()    
    Options.setWindowGray(false);
    Options.createGlobalLayer();
    window:setVisible(false)
end; 

function onButtonOk()
    if Options.activeCell then
        local action = Options.activeCell.data.action;
        local deviceDescription = action.device;
        local category = Options.form.containerControls.comboboxCategory:getText();
        local keyStr = window.controls.comboboxKeyButton:getText();
        local modifiersStr = window.controls.editboxAddedModifiers:getText();
        
        if ('' == keyStr) and ('' == modifiersStr) then
            show(false);
            return;
        end;
        
        if action.combos == nil then
            action.combos = {};
        end;
        
        local combos = action.combos;
        local newCombo = assembleCombo(keyStr, modifiersStr); --  
        local newComboStr = assembleComboString(newCombo); --    
        local prevAssignment, prevComboInd = checkAssignment(newCombo); --         ( )
        if (prevAssignment ~= nil) then
            for i,combo in base.pairs(prevAssignment.combos) do 
                if i == prevComboInd then
                    base.table.remove(prevAssignment.combos, i);--   
                    if prevAssignment.widget ~= nil then
                        Options.updateWidget(prevAssignment.widget, prevAssignment.combos);
                    end;
                    break;
                end;
            end;
        end;
        --base.print('newComboStr',newComboStr);
        for i,combo in base.ipairs(combos) do --      
            local str = assembleComboString(combo); --    
            --base.print('str',str);
            if str == newComboStr then --         
                show(false); --    
                return;
            end;
        end
        --        ,   
        base.table.insert(combos, newCombo);
        Options.updateWidget(Options.activeCell, combos);
        Options.modified = true;
    end;
    show(false);
end;

function assembleComboString(combo)
    local str = '';
    str = str .. combo.key
    if combo.reformers then
        for k = 1, #combo.reformers do
            local reformer = combo.reformers[k];
            str = str .. ' + ' .. reformer;
        end;
    end;
    return str;
end; 

function getReformersEqual(reformers1, reformers2)
  local result = true
  
  reformers1 = reformers1 or {}
  base.table.sort(reformers1)
  
  reformers2 = reformers2 or {}
  base.table.sort(reformers2)
  
  result = (#reformers1 == #reformers2)
  
  if result then
      for k, reformer1 in base.ipairs(reformers1) do 
          local reformer2 = reformers2[k]
          
          if reformer1 ~= reformer2 then
              result = false
              break
          end
      end
  end
  
  return result
end

function checkAssignment(newCombo)
    -- if nil == activeCell then
        -- return
    -- end;
    --U.traverseTable(newCombo);
    for k,v in base.pairs(Options.ACTIONS) do
        local ACTION = v;
        for i = 1,#ACTION do 
            local action = ACTION[i];
            local combos = action.combos or {};
            --base.print('action.name',action.name);
            for j = 1,#combos do 
                local combo = combos[j]
                --base.print('combo.key, newCombo.key', combo.key, newCombo.key);
                if (combo.key == newCombo.key) and (DEVICE == action.device)  then
                    if getReformersEqual(combo.reformers, newCombo.reformers) then -- records match exactly
                        return action, j;
                    end;
                end;
            end
        end
    end;
end;

function update()
    local keyStr = window.controls.comboboxKeyButton:getText();
    local modifiersStr = window.controls.editboxAddedModifiers:getText();
    --base.print('keyStr, modifiersStr',keyStr, modifiersStr);
    local newCombo = assembleCombo(keyStr, modifiersStr);
    local prevAssignment = checkAssignment(newCombo);
    window.controls.editboxCurrentSetting:setText('');
    if prevAssignment then
        --base.print('prevAssignment.name',prevAssignment.name);
        window.controls.editboxCurrentSetting:setText(prevAssignment.localizedName);
    end;
end;

function assembleCombo(keyStr, modifiersStr)
    local reformers = {};
    local i1,i2 = base.string.find(modifiersStr, '+');
    --base.print('modifiersStr',modifiersStr);
    while i1 ~= nil do
        local str = base.string.sub(modifiersStr, 1, i1 - 1);
        --base.print('str', str);
        modifiersStr = base.string.sub(modifiersStr, i2 + 1);
        --base.print('modifiersStr',modifiersStr);
        base.table.insert(reformers, str);
        i1,i2 = base.string.find(modifiersStr, '+');
    end;
    
    if modifiersStr ~= '' then
        base.table.insert(reformers, modifiersStr);
    end;
    base.table.sort(reformers);
    
    local combo = {
        key = keyStr,
        reformers = reformers,
    };
    return combo;
end; 

function onButtonReset()
    window.controls.comboboxKeyButton:setText('');
    window.controls.editboxAddedModifiers:setText('');
    window.controls.comboboxAddModifier:setText('');
    window.controls.editboxCurrentSetting:setText('');
end;

function onButtonCancel() 
    show(false);
end;

function onComboAddModifier(self, item)
    if item then
        local str = window.controls.editboxAddedModifiers:getText();
        local newText = item:getText();
        local i1,i2  = base.string.find(str, newText);
        if not i1 then
            if #str > 0 then
                window.controls.editboxAddedModifiers:setText(str .. ' + ' .. newText);
            else
                window.controls.editboxAddedModifiers:setText(newText);
            end;
        end;
    end;
    update();
end;


function onComboKeyButton()
    update();
end;

function createKeyLayer()
    local action = Options.activeCell.data.action;
    local device = action.device;
    if device.deviceType == Options.DEVICETYPES.mouse then
        return;
    end;
    local layerName = "keyLayer";
    NewInput.removeAllLayers();
    NewInput.clearReformers();
    NewInput.createLayer(layerName);
    for keyName, keyID in base.pairs(device.supportedKeys) do
        --base.print('keyName, keyID',keyName, keyID);
        NewInput.addKeyCombo(layerName, keyID, {}, keyID, 0, 0);
    end;
    for reformerID, reformerName in base.pairs(Options.ALL_REFORMERS_CODES) do
        --base.print('reformerID, reformerName',reformerID, reformerName);
        NewInput.addKeyCombo(layerName, reformerID, {}, reformerID, 0, 0);
    end;    
    NewInput.setTopLayer(layerName);
    base.setInputProcessor(onProcessInput);
    return layerName;
end; 

function createAxesLayer()
    local action = Options.activeCell.data.action;
    local device = action.device;
    local layerName = "axesLayer";
    axesParams = {}
    NewInput.removeAllLayers();
    NewInput.clearReformers();
    NewInput.createLayer(layerName);
    for axisName, axisID in base.pairs(device.supportedAxes) do
        --base.print('axisName, axisID',axisName, axisID);
        NewInput.addAxisCombo(layerName, axisID, {}, axisID);
    end;
    for reformerID, reformerName in base.pairs(Options.ALL_REFORMERS_CODES) do
        --base.print('reformerID, reformerName',reformerID, reformerName);
        NewInput.addKeyCombo(layerName, reformerID, {}, reformerID, 0, 0);
    end;    
    NewInput.setTopLayer(layerName);
    base.setInputProcessor(onProcessInput);
    return layerName;
end; 

function onProcessInput()
    --base.print('processing assignment');
    local action = Options.activeCell.data.action;
    local inputActions = NewInput.process();
    local category = Options.form.containerControls.comboboxCategory:getText();
    if #inputActions > 0 then
        for i = 1,#inputActions do 
            local inputAction = inputActions[i];
            --U.traverseTable(inputActions[i]);
            --U.traverseTable(action.device.keys);
            --local actionID = Options.INJOYSTICKN(action.device.joystickNumber,inputAction.action);
            local actionID = inputAction.action;
            local value = inputAction.value;
            --+-base.print('inputAction.device, actionID, value', inputAction.device, actionID, value);
            --local category = Options.form.containerControls.comboboxCategory:getText();
            local modifierName = Options.ALL_REFORMERS_CODES[actionID] or '';
            local axisName = action.device.axes[actionID] or '';
            local keyName = action.device.keys[actionID] or '';
            --base.print('actionID, keyName, axisName, modifierName',actionID, keyName, axisName, modifierName);
            if modifierName ~= '' then
                window.controls.comboboxAddModifier:setText(modifierName);
                local text = window.controls.editboxAddedModifiers:getText();
                local i1,i2 = base.string.find(text, modifierName);
                if not i1 then
                    local str;
                    if '' ~= text then
                        str = text .. '+' .. modifierName;
                    else
                        str = modifierName;
                    end;
                    window.controls.editboxAddedModifiers:setText(str);
                end;
            else
                local k = 0.5;
                local timeDelta = 1000;
                if cdata.axisCommands == category then
                    if not axesParams[actionID] then 
                        local axisData = {
                            minValue = nil;
                            maxValue = nil;
                            maxTime = 0;
                            minTime = 0;
                            valueMax = NewInput.getAxesValueMax();        
                        };
                        axesParams[actionID] = axisData;
                    end;
                    
                    local axisData = axesParams[actionID];
                    axisData.minValue = axisData.minValue or value;
                    if value < -k*axisData.valueMax then
                        --base.print('axisName passed lower bound',axisName, value , k*axisData.valueMax);
                        axisData.minValue = value;
                        axisData.minTime = inputAction.time;
                    end;
                    
                    axisData.maxValue = axisData.maxValue or value;
                    if value > k*axisData.valueMax then
                        --base.print('axisName passed upper bound',axisName,  value , k*axisData.valueMax);
                        axisData.maxValue = value;
                        axisData.maxTime = inputAction.time;
                    end;
                    
                    -- axisData.minValue = base.math.min(axisData.minValue or value, value);
                    -- axisData.maxValue = base.math.max(axisData.maxValue or value, value);
                    --local range = base.math.abs(axisData.maxValue - axisData.minValue);
                    --base.print('axisData.minValue, axisData.maxValue, range, 2*axisData.valueMax',axisData.minValue, axisData.maxValue, range, 2*axisData.valueMax);
                    local currTime = inputAction.time;
                    --base.print(range, k*2*axisData.valueMax, currTime - axisData.maxTime, currTime - axisData.minTime)
                    -- if (range >= k*2*axisData.valueMax) and
                        -- ( ( currTime - axisData.maxTime ) < 1000) and
                    if ( ( currTime - axisData.maxTime ) < timeDelta) 
                        and ( ( currTime - axisData.minTime ) < timeDelta)
                        -- ( ( currTime - axisData.minTime ) < 1000)
                        then
                        window.controls.comboboxKeyButton:setText(axisName);
                        base.print(axisName, actionID );
                    end;
                else
                    window.controls.comboboxKeyButton:setText(keyName);
                end;
            end;
        end
        update();
    end;
end; 